/*
	Interface to aclfs
*/
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <http_protocol.h>
#include "mod_dav.h"
#include <virtualfs/libaclfs.h>

struct apr_file_t {
	int fd;
};

struct apr_dir_t {
	DIR *dir;
	char path[PATH_MAX];
	char tmppath[PATH_MAX];
};

static void log (const char *ctl, ...)
{
	extern int dav_ufs_debug;
	if (dav_ufs_debug){
		FILE *f = fopen ("/var/log/dav_ufs.log","a");
		if (f != NULL){
			va_list list;
			va_start (list,ctl);
			vfprintf (f,ctl,list);
			va_end (list);
			fclose (f);
		}
	}
}

void afile_mapdir(const char *http_dir, const char *aclfs_volume)
{
	log ("mapdir :%s: :%s:\n",http_dir,aclfs_volume);
	aclfs_setmount (http_dir, aclfs_volume);
}

static apr_status_t afile_rstatus(int cond)
{
	apr_status_t ret = APR_SUCCESS;
	if (cond == 0){
		if (errno == 0){
			ret = EINVAL;
		}else{
			ret = errno;
		}
	}
	return ret;
}

apr_status_t afile_file_open (
	apr_file_t **inf,
	const char *f,
	apr_int32_t flags,
	apr_fileperms_t perm,
	apr_pool_t *p)
{
	int unix_flags = 0;
	int fd;
	if ((flags & (APR_READ|APR_WRITE))==(APR_READ|APR_WRITE)){
		unix_flags = O_RDWR;
	}else if (flags & APR_WRITE){
		unix_flags = O_WRONLY;
	}else if (flags & APR_READ){
		unix_flags = O_RDONLY;
	}
	if (flags & APR_CREATE) unix_flags |= O_CREAT;
	if (flags & APR_APPEND) unix_flags |= O_APPEND;
	if (flags & APR_TRUNCATE) unix_flags |= O_TRUNC;
	if (flags & APR_EXCL) unix_flags |= O_EXCL;
	fd = aclfs_open (f,unix_flags,perm);
	log ("open :%s: flags=%o,%o perm=%o fd=%d\n",f,flags,unix_flags,perm,fd);
	*inf = NULL;
	if (fd != -1){
		apr_file_t *f = apr_palloc(p,sizeof(apr_file_t));
		*inf = f;
		f->fd = fd;
	}
	return afile_rstatus(fd != -1);
}

apr_status_t afile_file_close (
	apr_file_t *inf)
{
	apr_status_t ret = afile_rstatus(aclfs_close (inf->fd)!=-1);
	return ret;
}

apr_status_t afile_file_remove (const char *dst, apr_pool_t *p)
{
	return afile_rstatus(aclfs_unlink (dst) != -1);
}

apr_status_t afile_file_write_full (
	apr_file_t *f,
	const void *buf,
	apr_size_t len,
	apr_size_t *written)
{
	int nb = write (f->fd,buf,len);
	log ("afile_file_write nbytes=%d nb=%d\n",len,nb);
	if (written != NULL) *written = nb;
	return afile_rstatus (nb == len);
}

apr_status_t afile_file_read_full(
	apr_file_t *f,
	void *buf,
	apr_size_t nbytes,
	apr_size_t *bytes_read)
{
	int nb = read (f->fd,buf,nbytes);
	log ("afile_file_read nbytes=%d nb=%d\n",nbytes,nb);
	if (bytes_read != NULL) *bytes_read = nb;
	return afile_rstatus (nb > 0);
}
apr_status_t afile_file_read(
	apr_file_t *thefile,
	void *buf,
	apr_size_t *nbytes)
{
	apr_size_t bytes_read;
	apr_status_t ret = afile_file_read_full (thefile,buf,*nbytes,&bytes_read);
	*nbytes = bytes_read;
	return ret;
}

apr_status_t afile_file_rename(
	const char *from_path, 
	const char *to_path,
	apr_pool_t *pool)
{
	return afile_rstatus(aclfs_rename(from_path,to_path)!=-1);
}

apr_status_t afile_file_seek(
	apr_file_t *f, 
	apr_seek_where_t where,
	apr_off_t *offset)
{
	log ("afile_file_seek %ld\n",where);
	return afile_rstatus (lseek(f->fd,where,SEEK_SET)!=-1);
}

apr_status_t afile_file_perms_set(
	const char *fname,
	apr_fileperms_t perms)
{
	log ("file_perms %s\n",fname);
	return afile_rstatus(aclfs_chmod (fname,perms)!=-1);
}

apr_status_t afile_dir_make(
	const char *path,
	apr_fileperms_t perm, 
	apr_pool_t *cont)
{
	log ("dir_make %s\n",path);
	return afile_rstatus(aclfs_mkdir (path,perm)!=-1);
}

apr_status_t afile_dir_remove(
	const char *path,
	apr_pool_t *cont)
{
	log ("dir_remove %s\n",path);
	return afile_rstatus(aclfs_rmdir (path)!=-1);
}


static void afile_fillinfo (
	apr_finfo_t *finfo,
	struct stat *st)
{
	if (S_ISREG(st->st_mode)){
		finfo->filetype = APR_REG;
	}else if (S_ISDIR(st->st_mode)){
		finfo->filetype = APR_DIR;
	}else if (S_ISLNK(st->st_mode)){
		finfo->filetype = APR_LNK;
	}
	finfo->size = st->st_size;
	finfo->csize = st->st_size;
	finfo->user = st->st_uid;
	finfo->group = st->st_gid;
	finfo->atime = st->st_atime;
	finfo->ctime = st->st_ctime;
	finfo->mtime = st->st_mtime;
	finfo->inode = st->st_ino;
	finfo->device = st->st_dev;
	finfo->protection = st->st_mode;
	finfo->valid = 0xffff;
}

apr_status_t afile_stat(
	apr_finfo_t *finfo,
	const char *fname,
	apr_int32_t wanted,
	apr_pool_t *cont)
{
	struct stat st;
	int ok = aclfs_stat(0,fname,&st);
	log ("afile_stat %s %d wanted=%x\n",fname,ok,wanted);
	if (ok != -1){
		afile_fillinfo (finfo,&st);
	}
	return afile_rstatus(ok != -1);
}
apr_status_t afile_lstat(
	apr_finfo_t *finfo,
	const char *fname,
	apr_int32_t wanted,
	apr_pool_t *cont)
{
	struct stat st;
	int ok = aclfs_lstat(0,fname,&st);
	log ("afile_lstat %s %d wanted=%x\n",fname,ok,wanted);
	if (ok != -1){
		afile_fillinfo (finfo,&st);
	}
	return afile_rstatus(ok != -1);

}

apr_status_t afile_dir_open(
	apr_dir_t **new_dir, 
	const char *dirname, 
	apr_pool_t *cont)
{
	DIR *dir = aclfs_opendir (dirname);
	log ("dir_open :%s: %p\n",dirname,dir);
	if (dir != NULL){
		apr_dir_t *t = apr_palloc(cont,sizeof(apr_dir_t));
		t->dir = dir;
		strncpy (t->path,dirname,sizeof(t->path)-1);
		*new_dir = t;
	}
	return afile_rstatus(dir != NULL);
}

apr_status_t afile_dir_close(apr_dir_t *thedir)
{
	aclfs_closedir (thedir->dir);
	return APR_SUCCESS;
}

apr_status_t afile_dir_read(
	apr_finfo_t *finfo,
	apr_int32_t wanted,
	apr_dir_t *f)
{
	struct dirent *ent = aclfs_readdir (f->dir);
	log ("dir_read ent=%p\n",ent);
	finfo->valid = 0;
	if (ent != NULL){
		log ("dir_read d_name = %s\n",ent->d_name);
		snprintf (f->tmppath,sizeof(f->tmppath)-1,"%s/%s",f->path,ent->d_name);
		finfo->valid = 0xffff;
		afile_lstat(finfo,f->tmppath,0xffffffff,NULL);
		finfo->fname = f->tmppath; // afile_strdup(finfo->pool,fpath);
		finfo->name = ent->d_name; // afile_strdup(finfo->pool,ent->d_name);
		finfo->filehand = NULL;
	}else{
		errno = ENOENT;
	}
	return afile_rstatus(ent != NULL);
}


apr_status_t afile_dir_rewind(apr_dir_t *f)
{
	aclfs_rewinddir (f->dir);
	return APR_SUCCESS;
}


int afile_authenticate_basic_user(request_rec *r)
{
	const char *sent_pw;
	int res;
	if ((res = ap_get_basic_auth_pw(r, &sent_pw))) {
		return res;
	}
	// log ("authenticate user=%s send_pw=%s\n",r->user,sent_pw);
	if (aclfs_setcred (r->user,sent_pw)!=-1){
		return OK;
	}
        return HTTP_UNAUTHORIZED;
}

/* Checking ID */

int afile_check_user_access(request_rec *r)
{
	log ("check_user_access user=%s\n",r->user);
	return OK;
#if 0
    auth_config_rec *conf = ap_get_module_config(r->per_dir_config,
                                                 &auth_module);
    char *user = r->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    apr_table_t *grpstatus;
    const apr_array_header_t *reqs_arr = ap_requires(r);
    require_line *reqs;

    /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive, 
     * then any user will do.
     */
    if (!reqs_arr) {
        return OK;
    }
    reqs = (require_line *)reqs_arr->elts;

    if (conf->auth_grpfile) {
        grpstatus = groups_for_user(r->pool, user, conf->auth_grpfile);
    }
    else {
        grpstatus = NULL;
    }

    for (x = 0; x < reqs_arr->nelts; x++) {

        if (!(reqs[x].method_mask & (AP_METHOD_BIT << m))) {
            continue;
        }

        method_restricted = 1;

        t = reqs[x].requirement;
        w = ap_getword_white(r->pool, &t);
        if (!strcmp(w, "valid-user")) {
            return OK;
        }
        if (!strcmp(w, "user")) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (!strcmp(user, w)) {
                    return OK;
                }
            }
        }
        else if (!strcmp(w, "group")) {
            if (!grpstatus) {
                return DECLINED;        /* DBM group?  Something else? */
            }

            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (apr_table_get(grpstatus, w)) {
                    return OK;
                }
            }
        }
        else if (conf->auth_authoritative) {
            /* if we aren't authoritative, any require directive could be
             * valid even if we don't grok it.  However, if we are 
             * authoritative, we can warn the user they did something wrong.
             * That something could be a missing "AuthAuthoritative off", but
             * more likely is a typo in the require directive.
             */
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "access to %s failed, reason: unknown require "
                          "directive:\"%s\"", r->uri, reqs[x].requirement);
        }
    }

    if (!method_restricted) {
        return OK;
    }

    if (!(conf->auth_authoritative)) {
        return DECLINED;
    }

    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "access to %s failed, reason: user %s not allowed access",
                  r->uri, user);
        
    ap_note_basic_auth_failure(r);
    return HTTP_UNAUTHORIZED;
#endif
}

void afile_sockpath (const char *path)
{
	aclfs_setsocket(path);
}

